import configparser
import datetime as dt
import logging
import os
import sys
import traceback
from glob import glob
from io import StringIO

import click
import iso8601
import numpy as np
from hspatial import NODATAVALUE
from htimeseries import HTimeseries, TzinfoFromString
from osgeo import gdal, ogr, osr

from . import __version__
from .evaporation import PenmanMonteith

gdal.UseExceptions()


class WrongValueError(configparser.Error):
    pass


class App:
    def __init__(self, configfilename):
        self.configfilename = configfilename

    def run(self):
        self.config = AppConfig(self.configfilename)
        self.config.read()
        self._setup_logger()
        self._execute_with_error_handling()

    def _execute_with_error_handling(self):
        self.logger.info("Starting evaporation, " + dt.datetime.today().isoformat())
        try:
            self._execute()
        except Exception as e:
            self.logger.error(str(e))
            self.logger.debug(traceback.format_exc())
            self.logger.info(
                "evaporation terminated with error, " + dt.datetime.today().isoformat()
            )
            raise click.ClickException(str(e))
        else:
            self.logger.info("Finished evaporation, " + dt.datetime.today().isoformat())

    def _setup_logger(self):
        self.logger = logging.getLogger("evaporation")
        self._set_logger_handler()
        self.logger.setLevel(self.config.loglevel.upper())

    def _set_logger_handler(self):
        if getattr(self.config, "logfile", None):
            self.logger.addHandler(logging.FileHandler(self.config.logfile))
        else:
            self.logger.addHandler(logging.StreamHandler())

    def _execute(self):
        if self.config.is_spatial:
            ProcessSpatial(self.config).execute()
        else:
            ProcessAtPoint(self.config).execute()


class AppConfig:
    config_file_options = {
        "logfile": {"fallback": ""},
        "loglevel": {"fallback": "warning"},
        "base_dir": {"fallback": ""},
        "step_length": {},
        "elevation": {"fallback": ""},
        "albedo": {},
        "nighttime_solar_radiation_ratio": {"fallback": 0.6},
        "unit_converter_temperature": {"fallback": "x"},
        "unit_converter_humidity": {"fallback": "x"},
        "unit_converter_wind_speed": {"fallback": "x"},
        "unit_converter_pressure": {"fallback": "x"},
        "unit_converter_solar_radiation": {"fallback": "x"},
        "temperature_min_prefix": {"fallback": "temperature_min"},
        "temperature_max_prefix": {"fallback": "temperature_max"},
        "temperature_prefix": {"fallback": "temperature"},
        "humidity_prefix": {"fallback": "humidity"},
        "humidity_min_prefix": {"fallback": "humidity_min"},
        "humidity_max_prefix": {"fallback": "humidity_max"},
        "wind_speed_prefix": {"fallback": "wind_speed"},
        "pressure_prefix": {"fallback": "pressure"},
        "solar_radiation_prefix": {"fallback": "solar_radiation"},
        "sunshine_duration_prefix": {"fallback": "sunshine_duration"},
        "evaporation_prefix": {"fallback": "evaporation"},
    }

    def __init__(self, configfilename):
        self.configfilename = configfilename

    def read(self):
        try:
            self._parse_config()
        except (OSError, configparser.Error) as e:
            sys.stderr.write(str(e))
            raise click.ClickException(str(e))

    def _parse_config(self):
        self._read_config_file()
        self._get_config_options()
        self._parse_config_options()
        self._parse_unit_converters()
        self._parse_elevation()

    def _read_config_file(self):
        self.config = configparser.ConfigParser(interpolation=None)
        try:
            self._read_config_file_assuming_it_has_section_headers()
        except configparser.MissingSectionHeaderError:
            self._read_config_file_without_sections()

    def _read_config_file_assuming_it_has_section_headers(self):
        with open(self.configfilename) as f:
            self.config.read_file(f)

    def _read_config_file_without_sections(self):
        with open(self.configfilename) as f:
            configuration = "[General]\n" + f.read()
        self.config.read_file(StringIO(configuration))

    def _get_config_options(self):
        self.options = {
            opt: self.config.get("General", opt, **kwargs)
            for opt, kwargs in self.config_file_options.items()
        }
        for key, value in self.options.items():
            setattr(self, key, value)

    def _parse_config_options(self):
        self._parse_log_level()
        self._parse_step_length()
        self._parse_unit_converters()
        self._set_is_spatial()
        self._parse_elevation()
        self._parse_albedo()
        self._parse_nighttime_solar_radiation_ratio()

    def _parse_log_level(self):
        log_levels = ("ERROR", "WARNING", "INFO", "DEBUG")
        self.loglevel = self.options["loglevel"].upper()
        if self.loglevel not in log_levels:
            raise WrongValueError("loglevel must be one of " + ", ".join(log_levels))

    def _parse_step_length(self):
        s = self.options["step_length"]
        self._check_step_length(s)
        self.step = dt.timedelta(minutes=int(s))

    def _check_step_length(self, s):
        if not s.isdecimal() or int(s) not in (60, 1440):
            raise WrongValueError(
                '"{}" is not an appropriate time step; in this version of '
                "vaporize, the step must be either 60 or 1440.".format(s)
            )

    def _parse_unit_converters(self):
        vars = {"temperature", "humidity", "wind_speed", "pressure", "solar_radiation"}
        self.unit_converters = {}
        for var in vars:
            config_item = "unit_converter_" + var
            config_value = self.options[config_item]
            lambda_defn = "lambda x: " + config_value
            self._set_unit_converter(var, config_item, config_value, lambda_defn)

    def _set_unit_converter(self, var, config_item, config_value, lambda_defn):
        try:
            self.unit_converters[var] = eval(lambda_defn)
        except Exception as e:
            raise WrongValueError(
                "{} while parsing {} ({}): {}".format(
                    e.__class__.__name__, config_item, config_value, str(e)
                )
            )

    def _parse_elevation(self):
        s = self.options["elevation"]
        if s == "":
            self.elevation = None
        else:
            self.elevation = self._get_number_or_grid(s)
        self._check_elevation()

    def _check_elevation(self):
        if self.is_spatial:
            self._check_elevation_for_spatial()
        else:
            self._check_elevation_for_point()

    def _check_elevation_for_spatial(self):
        if self.elevation is None:
            raise WrongValueError(
                "elevation needs to be specified in the configuration file"
            )
        elif np.any(self.elevation < -427) or np.any(self.elevation > 8848):
            raise WrongValueError("The elevation must be between -427 and 8848")

    def _check_elevation_for_point(self):
        if self.elevation is not None:
            raise WrongValueError(
                "elevation should only be specified in the configuration file "
                "in spatial calculation (otherwise we get it from the hts files)"
            )

    def _parse_albedo(self):
        s = self.options["albedo"].split()
        self._check_albedo_is_one_or_twelve_items(s)
        albedo = []
        for item in s:
            albedo.append(self._get_number_or_grid(item))
            self._check_albedo_domain(albedo[-1])
        self.albedo = albedo[0] if len(s) == 1 else albedo

    def _check_albedo_is_one_or_twelve_items(self, s):
        if len(s) not in (1, 12):
            raise ValueError(
                "Albedo must be either one item or 12 space-separated items"
            )

    def _check_albedo_domain(self, albedo):
        value_to_test = albedo if isinstance(albedo, float) else albedo.all()
        if value_to_test < 0.0 or value_to_test > 1.0:
            raise ValueError("Albedo must be between 0.0 and 1.0")

    def _parse_nighttime_solar_radiation_ratio(self):
        s = self.options["nighttime_solar_radiation_ratio"]
        self.nighttime_solar_radiation_ratio = self._get_number_or_grid(s)
        self._check_nighttime_solar_radiation_ratio()

    def _check_nighttime_solar_radiation_ratio(self):
        a = self.nighttime_solar_radiation_ratio
        if a < 0.4 or a > 0.8:
            raise WrongValueError(
                "The nighttime solar radiation ratio must " "be between 0.4 and 0.8"
            )

    def _get_number_or_grid(self, s):
        """Return either a number or a grid from s, depending on its contents.

        If string s holds a valid number, return it; otherwise try to open the
        geotiff file whose filename is s and read its first band into the
        returned numpy array.
        """
        try:
            return float(s)
        except ValueError:
            return self._get_grid(s)

    def _get_grid(self, s):
        input_file = gdal.Open(os.path.join(self.options["base_dir"], s))
        result = input_file.GetRasterBand(1).ReadAsArray()
        nodata = input_file.GetRasterBand(1).GetNoDataValue()
        if nodata is not None:
            result = np.ma.masked_values(result, nodata, copy=False)
        return result

    def _set_is_spatial(self):
        base_dir = self.options["base_dir"]
        base_dir_has_tif_files = bool(glob(os.path.join(base_dir, "*.tif")))
        base_dir_has_hts_files = bool(glob(os.path.join(base_dir, "*.hts")))
        self._check_tif_hts_consistency(base_dir_has_tif_files, base_dir_has_hts_files)
        self.is_spatial = base_dir_has_tif_files

    def _check_tif_hts_consistency(self, has_tif, has_hts):
        if has_tif and has_hts:
            raise WrongValueError(
                "Base directory {} contains both tif files and hts files; "
                "this is not allowed.".format(self.options["base_dir"])
            )
        elif not has_tif and not has_hts:
            raise WrongValueError(
                "Base directory {} contains neither tif files nor hts files.".format(
                    self.options["base_dir"]
                )
            )


class ProcessAtPoint:
    def __init__(self, config):
        self.config = config

    def execute(self):
        self._read_input_timeseries()
        self._setup_attrs()
        self._check_all_timeseries_are_in_same_location_and_timezone()
        self._get_location_in_wgs84()
        self._prepare_penman_monteith_parameters()
        self._prepare_resulting_htimeseries_object()
        self._determine_variables_to_use_in_calculation()
        self._calculate_evaporation()
        self._save_result()

    def _setup_attrs(self):
        atimeseries = self.input_timeseries["wind_speed"]
        self.location = atimeseries.location
        self.timezone = getattr(atimeseries, "timezone", None)

    def _read_input_timeseries(self):
        vars = {
            "temperature_min",
            "temperature_max",
            "temperature",
            "humidity",
            "humidity_min",
            "humidity_max",
            "wind_speed",
            "pressure",
            "solar_radiation",
            "sunshine_duration",
        }
        self.input_timeseries = {}
        for var in vars:
            self._get_input_timeseries_for_var(var)

    def _get_input_timeseries_for_var(self, var):
        filename = os.path.join(
            self.config.base_dir, getattr(self.config, var + "_prefix") + ".hts"
        )
        if not os.path.exists(filename):
            return
        with open(filename, "r", newline="\n") as f:
            self.input_timeseries[var] = HTimeseries(f)

    def _check_all_timeseries_are_in_same_location_and_timezone(self):
        for i, (name, hts) in enumerate(self.input_timeseries.items()):
            if i == 0:
                reference_hts = hts
            else:
                self._compare_location_and_timezone_of(hts, reference_hts)

    def _compare_location_and_timezone_of(self, hts1, hts2):
        self._compare_locations_of(hts1, hts2)
        self._compare_altitudes_of(hts1, hts2)
        self._compare_timezones_of(hts1, hts2)

    def _compare_locations_of(self, hts1, hts2):
        loc1 = hts1.location
        loc2 = hts2.location

        abscissas_differ = self._numbers_are_wrong_or_differ(
            loc1.get("abscissa"), loc2.get("abscissa")
        )
        ordinates_differ = self._numbers_are_wrong_or_differ(
            loc1.get("ordinate"), loc2.get("ordinate")
        )
        srids_differ = loc1.get("srid") != loc2.get("srid")

        if abscissas_differ or ordinates_differ or srids_differ:
            raise ValueError(
                "Incorrect or unspecified or inconsistent locations in the time series "
                "files."
            )

    def _numbers_are_wrong_or_differ(self, num1, num2, tolerance=1e7):
        if num1 is None or num2 is None:
            return True
        return abs(num1 - num2) > tolerance

    def _compare_altitudes_of(self, hts1, hts2):
        altitude1 = hts1.location.get("altitude")
        altitude2 = hts2.location.get("altitude")
        if self._numbers_are_wrong_or_differ(altitude1, altitude2, 1e-2):
            raise ValueError(
                "Incorrect or unspecified or inconsistent altitudes in the time series "
                "files."
            )

    def _compare_timezones_of(self, hts1, hts2):
        timezone1 = getattr(hts1, "timezone", "")
        timezone2 = getattr(hts2, "timezone", "")
        if timezone1 != timezone2:
            raise ValueError(
                "Incorrect or unspecified or inconsistent time zones in the time "
                "series files."
            )

    def _get_location_in_wgs84(self):
        source_projection = osr.SpatialReference()
        source_projection.ImportFromEPSG(self.location["srid"])
        wgs84 = osr.SpatialReference()
        wgs84.ImportFromEPSG(4326)
        transform = osr.CoordinateTransformation(source_projection, wgs84)
        apoint = ogr.Geometry(ogr.wkbPoint)
        apoint.AddPoint(self.location["abscissa"], self.location["ordinate"])
        apoint.Transform(transform)
        self.location["latitude"] = apoint.GetY()
        self.location["longitude"] = apoint.GetX()

    def _prepare_penman_monteith_parameters(self):
        nsrr = self.config.nighttime_solar_radiation_ratio
        self.penman_monteith = PenmanMonteith(
            albedo=self.config.albedo,
            nighttime_solar_radiation_ratio=nsrr,
            elevation=self.location["altitude"],
            latitude=self.location["latitude"],
            longitude=self.location["longitude"],
            step_length=self.config.step,
            unit_converters=self.config.unit_converters,
        )

    def _prepare_resulting_htimeseries_object(self):
        self.pet = HTimeseries()
        minutes = int(self.config.step.total_seconds() / 60)
        self.pet.time_step = str(minutes) + ",0"
        self.pet.unit = "mm"
        self.pet.timezone = self.timezone
        self.pet.variable = "Potential Evapotranspiration"
        self.pet.precision = 2 if self.config.step == dt.timedelta(hours=1) else 1
        self.pet.location = self.location

    def _determine_variables_to_use_in_calculation(self):
        if self.config.step == dt.timedelta(hours=1):
            vars = ["temperature", "humidity", "wind_speed", "solar_radiation"]
            if "pressure" in self.input_timeseries:
                vars.append("pressure")
        elif self.config.step == dt.timedelta(days=1):
            vars = (
                "temperature_max",
                "temperature_min",
                "humidity_max",
                "humidity_min",
                "wind_speed",
                (
                    "solar_radiation"
                    if "solar_radiation" in self.input_timeseries
                    else "sunshine_duration"
                ),
            )
        self.input_vars = vars

    def _calculate_evaporation(self):
        for adatetime in self.input_timeseries["wind_speed"].data.index:
            self._calculate_evaporation_for(adatetime)

    def _calculate_evaporation_for(self, adatetime):
        try:
            kwargs = {
                v: self.input_timeseries[v].data.loc[adatetime, "value"]
                for v in self.input_vars
            }
        except (IndexError, KeyError):
            return
        kwargs["adatetime"] = self._datetime64_to_aware_datetime(adatetime)
        self.pet.data.loc[adatetime, "value"] = self.penman_monteith.calculate(**kwargs)

    def _datetime64_to_aware_datetime(self, adatetime):
        result = adatetime.to_pydatetime()
        if self.timezone:
            result = result.replace(tzinfo=TzinfoFromString(self.timezone))
        return result

    def _save_result(self):
        outfilename = self.config.evaporation_prefix + ".hts"
        outpathname = os.path.join(self.config.base_dir, outfilename)
        with open(outpathname, "w") as f:
            self.pet.write(f, format=HTimeseries.FILE)


class ProcessSpatial:
    def __init__(self, config):
        self.config = config

    def execute(self):
        self._get_geographical_reference_file()
        self._get_timestamps()
        self._get_geodata()
        self._prepare_penman_monteith_parameters()
        self._calculate_evaporation_for_all_timestamps()
        self._remove_stale_evaporation_files()

    def _get_geographical_reference_file(self):
        # Arbitrarily use the first temperature file to extract location and
        # other geographical stuff. Elsewhere consistency of such data from all
        # other files with this file will be checked.
        pattern = os.path.join(
            self.config.base_dir, self.config.wind_speed_prefix + "-*.tif"
        )
        wind_speed_files = glob(pattern)
        self.geographical_reference_file = wind_speed_files[0]

    def _get_timestamps(self):
        pattern = self.config.wind_speed_prefix + "-*.tif"
        saved_cwd = os.getcwd()
        try:
            os.chdir(self.config.base_dir)
            wind_speed_files = glob(pattern)
        finally:
            os.chdir(saved_cwd)

        # Remove the prefix from the start and the .tif from the end, leaving
        # only the date.
        prefix_len = len(self.config.wind_speed_prefix)
        start = prefix_len + 1
        self.timestamps = [item[start:-4] for item in wind_speed_files]

    def _get_geodata(self):
        """
        Retrieve geographical stuff into self.latitude, self.longitude,
        self.width, self.height, self.geo_transform, self.projection. We do
        this by retrieving the data from self.geographical_reference_file.
        """
        # Read data from GeoTIFF file
        fp = gdal.Open(self.geographical_reference_file)
        self.width, self.height = fp.RasterXSize, fp.RasterYSize
        self.geo_transform = fp.GetGeoTransform()
        self.projection = osr.SpatialReference()
        self.projection.ImportFromWkt(fp.GetProjection())

        # Find (x_left, y_top), (x_right, y_bottom)
        x_left, x_step, d1, y_top, d2, y_step = self.geo_transform
        x_right = x_left + self.width * x_step
        y_bottom = y_top + self.height * y_step

        # Transform into (long_left, lat_top), (long_right, lat_bottom)
        wgs84 = osr.SpatialReference()
        wgs84.ImportFromEPSG(4326)
        transform = osr.CoordinateTransformation(self.projection, wgs84)
        top_left = ogr.Geometry(ogr.wkbPoint)
        top_left.AddPoint(x_left, y_top)
        bottom_right = ogr.Geometry(ogr.wkbPoint)
        bottom_right.AddPoint(x_right, y_bottom)
        top_left.Transform(transform)
        bottom_right.Transform(transform)
        long_left, lat_top = top_left.GetX(), top_left.GetY()
        long_right, lat_bottom = bottom_right.GetX(), bottom_right.GetY()

        # Calculate self.latitude and self.longitude
        long_step = (long_right - long_left) / self.width
        longitudes = np.arange(
            long_left + long_step / 2.0, long_left + long_step * self.width, long_step
        )
        lat_step = (lat_top - lat_bottom) / self.height
        latitudes = np.arange(
            lat_top + lat_step / 2.0, lat_top + lat_step * self.height, lat_step
        )
        self.longitude, self.latitude = np.meshgrid(longitudes, latitudes)

    def _prepare_penman_monteith_parameters(self):
        nsrr = self.config.nighttime_solar_radiation_ratio
        self.penman_monteith = PenmanMonteith(
            albedo=self.config.albedo,
            nighttime_solar_radiation_ratio=nsrr,
            elevation=self.config.elevation,
            latitude=self.latitude,
            longitude=self.longitude,
            step_length=self.config.step,
            unit_converters=self.config.unit_converters,
        )

    def _calculate_evaporation_for_all_timestamps(self):
        for timestamp in self.timestamps:
            self._calculate_evaporation_for(timestamp)

    def _calculate_evaporation_for(self, timestamp):
        self._read_input_geofiles(timestamp)
        self._verify_that_solar_radiation_or_sunshine_duration_was_present(timestamp)
        adatetime = self._get_adatetime_arg(timestamp)
        result = self.penman_monteith.calculate(adatetime=adatetime, **self.input_data)
        self._write_result_to_file(result, timestamp, adatetime)

    def _read_input_geofiles(self, timestamp):
        self.input_data = {v: None for v in self._get_variables()}
        for var in self.input_data:
            self._read_input_geofile(var, timestamp)

    def _get_variables(self):
        if self.config.step == dt.timedelta(minutes=60):
            return {
                "temperature",
                "humidity",
                "wind_speed",
                "pressure",
                "solar_radiation",
            }
        else:
            return {
                "temperature_max",
                "temperature_min",
                "humidity_max",
                "humidity_min",
                "wind_speed",
                "solar_radiation",
                "sunshine_duration",
            }

    def _read_input_geofile(self, variable, timestamp):
        fp = self._open_geofile(variable, timestamp)
        if fp is None:
            return
        self._verify_consistency_of_geodata(fp)
        self.input_data[variable] = self._read_array_from_geofile(fp)
        fp = None

    def _open_geofile(self, variable, timestamp):
        filename_prefix = getattr(self.config, variable + "_prefix")
        filename = filename_prefix + "-" + timestamp + ".tif"
        self._filename = os.path.join(self.config.base_dir, filename)
        if variable in ("solar_radiation", "sunshine_duration"):
            if not os.path.exists(self._filename):
                # Either solar_radiation or sunshine_duration may be absent; here we
                # allow both to be absent and elsewhere we will check that one was
                # present
                return
        try:
            return gdal.Open(self._filename)
        except RuntimeError:
            if variable == "pressure":
                return
            raise

    def _verify_consistency_of_geodata(self, fp):
        consistent = all(
            (
                self.width == fp.RasterXSize,
                self.height == fp.RasterYSize,
                self.geo_transform == fp.GetGeoTransform(),
                self.projection.ExportToWkt() == fp.GetProjection(),
            )
        )
        if not consistent:
            raise Exception(
                "Not all input files have the same "
                "width, height, geo_transform and projection "
                "(offending items: {} and {})".format(
                    self.geographical_reference_file, self._filename
                )
            )

    def _read_array_from_geofile(self, fp):
        array = fp.GetRasterBand(1).ReadAsArray()
        nodata = fp.GetRasterBand(1).GetNoDataValue()
        if nodata is not None:
            array = np.ma.masked_values(array, nodata, copy=False)
        return array

    def _verify_that_solar_radiation_or_sunshine_duration_was_present(self, timestamp):
        if self.input_data["solar_radiation"] is not None:
            self.input_data.pop("sunshine_duration", None)
        elif self.input_data.get("sunshine_duration", None) is not None:
            self.input_data.pop("solar_radiation", None)
        else:
            raise RuntimeError(
                "Neither sunshine_duration nor solar_radiation are available for "
                + timestamp
            )

    def _get_adatetime_arg(self, timestamp):
        adatetime = iso8601.parse_date(
            self._timestamp_from_filename(timestamp), default_timezone=None
        )
        if self.config.step == dt.timedelta(minutes=1440):
            adatetime = adatetime.date()
        elif adatetime.tzinfo is None:
            raise Exception(
                "The time stamp in the input files does not have a time zone specified."
            )
        return adatetime

    def _timestamp_from_filename(self, s):
        """Convert a timestamp from its filename format  to its iso format

        E.g. from 2014-10-01-15-00-0100 to 2014-10-01 15:00-0100).
        """
        first_hyphen = s.find("-")
        if first_hyphen < 0:
            return s
        second_hyphen = s.find("-", first_hyphen + 1)
        if second_hyphen < 0:
            return s
        third_hyphen = s.find("-", second_hyphen + 1)
        if third_hyphen < 0:
            return s
        fourth_hyphen = s.find("-", third_hyphen + 1)
        chars = list(s)
        chars[third_hyphen] = " "
        if fourth_hyphen > 0:
            chars[fourth_hyphen] = ":"
        return "".join(chars)

    def _write_result_to_file(self, result, timestamp, adatetime):
        output_filename = self.config.evaporation_prefix + "-" + timestamp + ".tif"
        output_pathname = os.path.join(self.config.base_dir, output_filename)
        output = gdal.GetDriverByName("GTiff").Create(
            output_pathname, self.width, self.height, 1, gdal.GDT_Float32
        )
        try:
            output.SetMetadataItem("TIMESTAMP", adatetime.isoformat())
            output.SetGeoTransform(self.geo_transform)
            output.SetProjection(self.projection.ExportToWkt())
            result[result.mask] = NODATAVALUE
            output.GetRasterBand(1).SetNoDataValue(NODATAVALUE)
            output.GetRasterBand(1).WriteArray(result)
        finally:
            # Close the dataset
            output = None

    def _remove_stale_evaporation_files(self):
        """Remove evaporation files for which no input files exist.
        """
        pattern = self.config.evaporation_prefix + "-*.tif"
        saved_cwd = os.getcwd()
        try:
            os.chdir(self.config.base_dir)
            evaporation_files = glob(pattern)
            prefix_len = len(self.config.evaporation_prefix)
            for filename in evaporation_files:
                start = prefix_len + 1
                if filename[start:-4] not in self.timestamps:
                    os.unlink(filename)
        finally:
            os.chdir(saved_cwd)


@click.command()
@click.argument("configfile")
@click.version_option(version=__version__, prog_name="vaporize")
def main(configfile):
    """Calculation of evaporation and transpiration"""

    app = App(configfile)
    app.run()



import math
import warnings
from datetime import timedelta
from math import cos, pi, sin, tan

import numpy as np

# Note about RuntimeWarning
#
# When numpy makes calculations with masked arrays, it sometimes emits spurious
# RuntimeWarnings. This is because it occasionally does use the masked part of
# the array during the calculations (but masks the result). This is a known
# numpy bug (e.g. https://github.com/numpy/numpy/issues/4269). The numpy
# documentation, section "Operations on masked arrays", also has a related
# warning there.
#
# In order to avoid these spurious warnings, we have used, at various places in
# the code, "with warnings.catch_warnings()". We have attempted to unit test
# it, but sometimes it's hard to make the bug appear. A large array in
# production may cause the bug, but a small array in the unit test might not
# cause it, despite same python and numpy version. So the locations in which
# a fix was needed were largely located in production.


class PenmanMonteith(object):
    # Stefan-Boltzmann constant (Allen et al., 1998, p. 52)
    sigma = 4.903e-9

    def __init__(
        self,
        albedo,
        elevation,
        latitude,
        step_length,
        longitude=None,
        nighttime_solar_radiation_ratio=None,
        unit_converters={},
    ):
        self.albedo = albedo
        self.nighttime_solar_radiation_ratio = nighttime_solar_radiation_ratio
        self.elevation = elevation
        self.latitude = latitude
        self.longitude = longitude
        self.step_length = step_length
        self.unit_converters = unit_converters

    def calculate(self, **kwargs):
        if self.step_length == timedelta(minutes=60):
            return self.calculate_hourly(**kwargs)
        elif self.step_length == timedelta(days=1):
            return self.calculate_daily(**kwargs)
        else:
            raise NotImplementedError(
                "Evaporation for time steps other than hourly and daily "
                "has not been implemented."
            )

    def calculate_daily(
        self,
        temperature_max,
        temperature_min,
        humidity_max,
        humidity_min,
        wind_speed,
        adatetime,
        sunshine_duration=None,
        pressure=None,
        solar_radiation=None,
    ):
        if pressure is None:
            # Eq. 7 p. 31
            pressure = 101.3 * ((293 - 0.0065 * self.elevation) / 293) ** 5.26
        variables = self.convert_units(
            temperature_max=temperature_max,
            temperature_min=temperature_min,
            humidity_max=humidity_max,
            humidity_min=humidity_min,
            wind_speed=wind_speed,
            sunshine_duration=sunshine_duration,
            pressure=pressure,
        )

        # Radiation
        r_a, N = self.get_extraterrestrial_radiation(adatetime)
        if solar_radiation is None:
            solar_radiation = (
                0.25 + 0.50 * variables["sunshine_duration"] / N
            ) * r_a  # Eq.35 p. 50
        r_so = r_a * (0.75 + 2e-5 * self.elevation)  # Eq. 37, p. 51
        variables.update(self.convert_units(solar_radiation=solar_radiation))

        with warnings.catch_warnings():
            # See comment about RuntimeWarning on top of the file
            warnings.simplefilter("ignore", RuntimeWarning)
            temperature_mean = (
                variables["temperature_max"] + variables["temperature_min"]
            ) / 2
        variables["temperature_mean"] = temperature_mean
        gamma = self.get_psychrometric_constant(temperature_mean, pressure)
        return self.penman_monteith_daily(
            incoming_solar_radiation=variables["solar_radiation"],
            clear_sky_solar_radiation=r_so,
            psychrometric_constant=gamma,
            mean_wind_speed=variables["wind_speed"],
            temperature_max=variables["temperature_max"],
            temperature_min=variables["temperature_min"],
            temperature_mean=variables["temperature_mean"],
            humidity_max=variables["humidity_max"],
            humidity_min=variables["humidity_min"],
            adate=adatetime,
        )

    def calculate_hourly(
        self,
        temperature,
        humidity,
        wind_speed,
        solar_radiation,
        adatetime,
        pressure=None,
    ):
        if pressure is None:
            # Eq. 7 p. 31
            pressure = 101.3 * ((293 - 0.0065 * self.elevation) / 293) ** 5.26
        variables = self.convert_units(
            temperature=temperature,
            humidity=humidity,
            wind_speed=wind_speed,
            pressure=pressure,
            solar_radiation=solar_radiation,
        )
        gamma = self.get_psychrometric_constant(
            variables["temperature"], variables["pressure"]
        )
        r_so = self.get_extraterrestrial_radiation(adatetime) * (
            0.75 + 2e-5 * self.elevation
        )  # Eq. 37, p. 51
        return self.penman_monteith_hourly(
            incoming_solar_radiation=variables["solar_radiation"],
            clear_sky_solar_radiation=r_so,
            psychrometric_constant=gamma,
            mean_wind_speed=variables["wind_speed"],
            mean_temperature=variables["temperature"],
            mean_relative_humidity=variables["humidity"],
            adatetime=adatetime,
        )

    def convert_units(self, **kwargs):
        result = {}
        for item in kwargs:
            varname = item
            if item.endswith("_max") or item.endswith("_min"):
                varname = item[:-4]
            converter = self.unit_converters.get(varname, lambda x: x)
            with warnings.catch_warnings():
                # See comment about RuntimeWarning on top of the file
                warnings.simplefilter("ignore", RuntimeWarning)
                result[item] = converter(kwargs[item])
        return result

    def get_extraterrestrial_radiation(self, adatetime):
        """
        Calculates the solar radiation we would receive if there were no
        atmosphere. This is a function of date, time and location.

        If adatetime is a datetime object, it merely returns the
        extraterrestrial radiation R_a; if it is a date object, it returns a
        tuple, (R_a, N), where N is the daylight hours.
        """
        j = adatetime.timetuple().tm_yday  # Day of year

        # Inverse relative distance Earth-Sun, eq. 23, p. 46.
        dr = 1 + 0.033 * cos(2 * pi * j / 365)

        # Solar declination, eq. 24, p. 46.
        decl = 0.409 * sin(2 * pi * j / 365 - 1.39)

        if self.step_length > timedelta(minutes=60):  # Daily?
            phi = self.latitude / 180.0 * pi
            omega_s = np.arccos(-np.tan(phi) * tan(decl))  # Eq. 25 p. 46

            r_a = (
                24
                * 60
                / pi
                * 0.0820
                * dr
                * (
                    omega_s * np.sin(phi) * sin(decl)
                    + np.cos(phi) * cos(decl) * np.sin(omega_s)
                )
            )  # Eq. 21 p. 46
            n = 24 / pi * omega_s  # Eq. 34 p. 48
            return r_a, n

        # Seasonal correction for solar time, eq. 32, p. 48.
        b = 2 * pi * (j - 81) / 364
        sc = 0.1645 * sin(2 * b) - 0.1255 * cos(b) - 0.025 * sin(b)

        # Longitude at the centre of the local time zone
        utc_offset = adatetime.utcoffset()
        utc_offset_hours = utc_offset.days * 24 + utc_offset.seconds / 3600.0
        lz = -utc_offset_hours * 15

        # Solar time angle at midpoint of the time period, eq. 31, p. 48.
        tm = adatetime - self.step_length / 2
        t = tm.hour + tm.minute / 60.0
        omega = pi / 12 * ((t + 0.06667 * (lz + self.longitude) + sc) - 12)

        # Solar time angles at beginning and end of the period, eqs. 29 and 30,
        # p. 48.
        t1 = self.step_length.seconds / 3600.0
        omega1 = omega - pi * t1 / 24
        omega2 = omega + pi * t1 / 24

        # Result: eq. 28, p. 47.
        phi = self.latitude / 180.0 * pi
        return (
            12
            * 60
            / pi
            * 0.0820
            * dr
            * (
                (omega2 - omega1) * np.sin(phi) * sin(decl)
                + np.cos(phi) * cos(decl) * (np.sin(omega2) - np.sin(omega1))
            )
        )

    def get_psychrometric_constant(self, temperature, pressure):
        """
        Allen et al. (1998), eq. 8, p. 32.

        This is called a "constant" because, although it is a function of
        temperature and pressure, its variations are small, and therefore it
        can be assumed constant for a location assuming standard pressure at
        that elevation and 20 degrees C. However, here we actually calculate
        it, so it isn't a constant.
        """
        lambda_ = 2.501 - (2.361e-3) * temperature  # eq. 3-1, p. 223
        return 1.013e-3 * pressure / 0.622 / lambda_

    def penman_monteith_daily(
        self,
        incoming_solar_radiation,
        clear_sky_solar_radiation,
        psychrometric_constant,
        mean_wind_speed,
        temperature_max,
        temperature_min,
        temperature_mean,
        humidity_max,
        humidity_min,
        adate,
    ):
        """
        Calculates and returns the reference evapotranspiration according
        to Allen et al. (1998), eq. 6, p. 24 & 65.
        """

        # Saturation and actual vapour pressure
        svp_max = self.get_saturation_vapour_pressure(temperature_max)
        svp_min = self.get_saturation_vapour_pressure(temperature_min)
        avp1 = svp_max * humidity_min / 100
        avp2 = svp_min * humidity_max / 100
        svp = (svp_max + svp_min) / 2  # Eq. 12 p. 36
        avp = (avp1 + avp2) / 2  # Eq. 12 p. 36

        # Saturation vapour pressure curve slope
        delta = self.get_saturation_vapour_pressure_curve_slope(temperature_mean)

        # Net incoming radiation; p. 51, eq. 38
        albedo = (
            self.albedo[adate.month - 1]
            if self.albedo.__class__.__name__ in ("tuple", "list")
            else self.albedo
        )
        rns = (1.0 - albedo) * incoming_solar_radiation

        # Net outgoing radiation
        rnl = self.get_net_outgoing_radiation(
            (temperature_min, temperature_max),
            incoming_solar_radiation,
            clear_sky_solar_radiation,
            avp,
        )

        # Net radiation at grass surface
        rn = rns - rnl

        # Soil heat flux
        g_day = 0  # Eq. 42 p. 54

        # Apply the formula
        numerator_term1 = 0.408 * delta * (rn - g_day)
        numerator_term2 = (
            psychrometric_constant
            * 900
            / (temperature_mean + 273.16)
            * mean_wind_speed
            * (svp - avp)
        )
        denominator = delta + psychrometric_constant * (1 + 0.34 * mean_wind_speed)

        return (numerator_term1 + numerator_term2) / denominator

    def penman_monteith_hourly(
        self,
        incoming_solar_radiation,
        clear_sky_solar_radiation,
        psychrometric_constant,
        mean_wind_speed,
        mean_temperature,
        mean_relative_humidity,
        adatetime,
    ):
        """
        Calculates and returns the reference evapotranspiration according
        to Allen et al. (1998), eq. 53, p. 74.

        As explained in Allen et al. (1998, p. 74), the function is
        modified in relation to the original Penman-Monteith equation, so
        that it is suitable for hourly data.
        """

        # Saturation and actual vapour pressure
        svp = self.get_saturation_vapour_pressure(mean_temperature)
        with warnings.catch_warnings():
            # See comment about RuntimeWarning on top of the file
            warnings.simplefilter("ignore", RuntimeWarning)
            avp = svp * mean_relative_humidity / 100.0  # Eq. 54, p. 74

        # Net incoming radiation; p. 51, eq. 38
        albedo = (
            self.albedo[adatetime.month - 1]
            if self.albedo.__class__.__name__ in ("tuple", "list")
            else self.albedo
        )
        rns = (1.0 - albedo) * incoming_solar_radiation

        # Net outgoing radiation
        rnl = self.get_net_outgoing_radiation(
            mean_temperature, incoming_solar_radiation, clear_sky_solar_radiation, avp
        )

        # Net radiation at grass surface
        rn = rns - rnl

        # Saturation vapour pressure curve slope
        delta = self.get_saturation_vapour_pressure_curve_slope(mean_temperature)

        # Soil heat flux density
        g = self.get_soil_heat_flux_density(incoming_solar_radiation, rn)

        # Apply the formula
        numerator_term1 = 0.408 * delta * (rn - g)
        with warnings.catch_warnings():
            # See comment about RuntimeWarning on top of the file
            warnings.simplefilter("ignore", RuntimeWarning)
            numerator_term2 = (
                psychrometric_constant
                * 37
                / (mean_temperature + 273.16)
                * mean_wind_speed
                * (svp - avp)
            )
        denominator = delta + psychrometric_constant * (1 + 0.34 * mean_wind_speed)

        return (numerator_term1 + numerator_term2) / denominator

    def get_net_outgoing_radiation(
        self,
        temperature,
        incoming_solar_radiation,
        clear_sky_solar_radiation,
        mean_actual_vapour_pressure,
    ):
        """
        Allen et al. (1998), p. 52, eq. 39. Temperature can be a tuple (a pair)
        of min and max, or a single value. If it is a single value, the
        equation is modified according to end of page 74.
        """
        if temperature.__class__.__name__ in ("tuple", "list"):
            with warnings.catch_warnings():
                # See comment about RuntimeWarning on top of the file
                warnings.simplefilter("ignore", RuntimeWarning)
                factor1 = (
                    self.sigma
                    * ((temperature[0] + 273.16) ** 4 + (temperature[1] + 273.16) ** 4)
                    / 2
                )
        else:
            with warnings.catch_warnings():
                # See comment about RuntimeWarning on top of the file
                warnings.simplefilter("ignore", RuntimeWarning)
                factor1 = self.sigma / 24 * (temperature + 273.16) ** 4
        factor2 = 0.34 - 0.14 * (mean_actual_vapour_pressure ** 0.5)

        # Solar radiation ratio Rs/Rs0 (Allen et al., 1998, top of p. 75).
        with warnings.catch_warnings():
            # See comment about RuntimeWarning on top of the file
            warnings.simplefilter("ignore", RuntimeWarning)
            solar_radiation_ratio = np.where(
                clear_sky_solar_radiation > 0.05,
                incoming_solar_radiation / clear_sky_solar_radiation,
                self.nighttime_solar_radiation_ratio,
            )
        solar_radiation_ratio = np.maximum(solar_radiation_ratio, 0.3)
        solar_radiation_ratio = np.minimum(solar_radiation_ratio, 1.0)

        factor3 = 1.35 * solar_radiation_ratio - 0.35

        return factor1 * factor2 * factor3

    def get_saturation_vapour_pressure(self, temperature):
        "Allen et al. (1998), p. 36, eq. 11."
        with warnings.catch_warnings():
            # See comment about RuntimeWarning on top of the file
            warnings.simplefilter("ignore")
            return 0.6108 * math.e ** (17.27 * temperature / (237.3 + temperature))

    def get_soil_heat_flux_density(self, incoming_solar_radiation, rn):
        "Allen et al. (1998), p. 55, eq. 45 & 46."
        coefficient = np.where(incoming_solar_radiation > 0.05, 0.1, 0.5)
        return coefficient * rn

    def get_saturation_vapour_pressure_curve_slope(self, temperature):
        "Allen et al. (1998), p. 37, eq. 13."
        numerator = 4098 * self.get_saturation_vapour_pressure(temperature)
        with warnings.catch_warnings():
            # See comment about RuntimeWarning on top of the file
            warnings.simplefilter("ignore", RuntimeWarning)
            denominator = (temperature + 237.3) ** 2
            return numerator / denominator



import datetime as dt
import os
import shutil
import sys
import tempfile
import textwrap
from io import StringIO
from pathlib import Path
from unittest import TestCase
from unittest.mock import patch

import click
import numpy as np
import pandas as pd
from click.testing import CliRunner
from htimeseries import HTimeseries
from osgeo import gdal, osr

from evaporation import cli

from .test_evaporation import senegal_tzinfo

gdal.UseExceptions()


def create_geotiff_file(filename, value):
    geo_transform = (-16.25, 1.0, 0, 16.217, 0, 1.0)
    wgs84 = osr.SpatialReference()
    wgs84.ImportFromEPSG(4326)
    f = gdal.GetDriverByName("GTiff").Create(filename, 2, 1, 1, gdal.GDT_Float32)
    try:
        f.SetGeoTransform(geo_transform)
        f.SetProjection(wgs84.ExportToWkt())
        f.GetRasterBand(1).WriteArray(value)
    finally:
        # Close the dataset
        f = None


class NonExistentConfigFileTestCase(TestCase):
    def setUp(self):
        runner = CliRunner(mix_stderr=False)
        self.result = runner.invoke(cli.main, ["nonexistent.conf"])

    def test_exit_status(self):
        self.assertEqual(self.result.exit_code, 1)

    def test_error_message(self):
        self.assertIn(
            "No such file or directory: 'nonexistent.conf'", self.result.stderr
        )


class WrongPointConfigurationTestCase(TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.configfilename = os.path.join(self.tempdir, "evaporation.conf")
        htsfilename = os.path.join(self.tempdir, "wind_speed.hts")
        Path(htsfilename).touch()

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def test_nonexistent_log_level_raises_error(self):
        with open(self.configfilename, "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                    base_dir = {self.tempdir}
                    albedo = 0.23
                    nighttime_solar_radiation_ratio = 0.8
                    elevation = 8
                    step_length = 60
                    unit_converter_pressure = x / 10.0
                    unit_converter_solar_radiation = x * 3600 /1e6
                    loglevel = NONEXISTENT_LOG_LEVEL
                    """.format(
                        self=self
                    )
                )
            )
        msg = "loglevel must be one of ERROR, WARNING, INFO, DEBUG"
        with self.assertRaisesRegex(click.ClickException, msg):
            cli.App(self.configfilename).run()

    def test_missing_step_raises_error(self):
        with open(self.configfilename, "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                    base_dir = {self.tempdir}
                    albedo = 0.23
                    nighttime_solar_radiation_ratio = 0.8
                    elevation = 8
                    """
                ).format(self=self)
            )
        with self.assertRaisesRegex(click.ClickException, "step_length"):
            cli.App(self.configfilename).run()

    def test_missing_albedo_raises_error(self):
        with open(self.configfilename, "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                    base_dir = {self.tempdir}
                    nighttime_solar_radiation_ratio = 0.8
                    elevation = 8
                    step_length = 60
                    """
                ).format(self=self)
            )
        with self.assertRaisesRegex(click.ClickException, "albedo"):
            cli.App(self.configfilename).run()


class WrongSpatialConfigurationTestCase(TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.configfilename = os.path.join(self.tempdir, "evaporation.conf")
        htsfilename = os.path.join(self.tempdir, "wind_speed.tif")
        Path(htsfilename).touch()

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def test_missing_elevation_raises_error(self):
        with open(self.configfilename, "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                base_dir = {self.tempdir}
                albedo = 0.23
                nighttime_solar_radiation_ratio = 0.8
                step_length = 60
                """
                ).format(self=self)
            )
        with self.assertRaisesRegex(click.ClickException, "elevation"):
            cli.App(self.configfilename).run()

    def test_single_albedo_with_wrong_domain_float_inputs(self):
        with open(self.configfilename, "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                base_dir = {self.tempdir}
                step_length = 60
                albedo = 2
                nighttime_solar_radiation_ratio = 0.8
                elevation = 8
                """
                ).format(self=self)
            )
        with self.assertRaisesRegex(ValueError, "Albedo must be between 0.0 and 1.0"):
            cli.App(self.configfilename).run()

    def test_seasonal_albedo_configuration_with_not_enough_arguments(self):
        with open(self.configfilename, "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                base_dir = {self.tempdir}
                step_length = 60
                albedo = a01.tiff a02.tiff a11.tiff a12.tiff
                nighttime_solar_radiation_ratio = 0.8
                elevation = 8
                """
                ).format(self=self)
            )
        msg = "Albedo must be either one item or 12 space-separated items"
        with self.assertRaisesRegex(ValueError, msg):
            cli.App(self.configfilename).run()

    def test_seasonal_albedo_with_wrong_domain_mixin_inputs(self):
        with open(self.configfilename, "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                base_dir = {self.tempdir}
                step_length = 60
                albedo = 1 2 0.34 0.24 0.45 0.4
                        0.34 0.12 a00.tif 0.78 0.78 2
                nighttime_solar_radiation_ratio = 0.8
                elevation = 8
                """
                ).format(self=self)
            )
        with self.assertRaisesRegex(ValueError, "Albedo must be between 0.0 and 1.0"):
            cli.App(self.configfilename).run()


class CorrectPointConfigurationTestCase(TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.configfilename = os.path.join(self.tempdir, "evaporation.conf")
        htsfilename = os.path.join(self.tempdir, "wind_speed.hts")
        Path(htsfilename).touch()

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    @patch("evaporation.cli.ProcessAtPoint")
    def test_executes(self, m):
        with open(self.configfilename, "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                    base_dir = {self.tempdir}
                    albedo = 0.23
                    nighttime_solar_radiation_ratio = 0.8
                    step_length = 60
                    unit_converter_pressure = x / 10.0
                    unit_converter_solar_radiation = x * 3600 /1e6
                    """
                ).format(self=self)
            )
        cli.App(self.configfilename).run()
        m.return_value.execute.assert_called_once_with()

    @patch("evaporation.cli.ProcessAtPoint")
    def test_albedo_configuration_as_one_number(self, m):
        with open(self.configfilename, "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                base_dir = {self.tempdir}
                step_length = 60
                albedo = 0.23
                nighttime_solar_radiation_ratio = 0.8
                """
                ).format(self=self)
            )
        cli.App(self.configfilename).run()
        m.return_value.execute.assert_called_once_with()


class CorrectSpatialConfigurationTestCase(TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.configfilename = os.path.join(self.tempdir, "evaporation.conf")
        htsfilename = os.path.join(self.tempdir, "wind_speed.tif")
        Path(htsfilename).touch()
        self._create_albedo_files()

    def _create_albedo_files(self):
        items = (
            "00",
            "01",
            "02",
            "03",
            "04",
            "05",
            "06",
            "07",
            "08",
            "09",
            "10",
            "11",
            "12",
        )
        for item in items:
            filename = os.path.join(self.tempdir, "a{}.tif".format(item))
            create_geotiff_file(filename, np.array([[0.23, 0.44]]))

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    @patch("evaporation.cli.ProcessSpatial")
    def test_albedo_configuration_as_one_grid(self, m):
        with open(self.configfilename, "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                base_dir = {self.tempdir}
                step_length = 60
                albedo = a00.tif
                nighttime_solar_radiation_ratio = 0.8
                elevation = 8
                """
                ).format(self=self)
            )
        cli.App(self.configfilename).run()
        m.return_value.execute.assert_called_once_with()

    @patch("evaporation.cli.ProcessSpatial")
    def test_seasonal_albedo_configuration_as_12_grids(self, m):
        with open(self.configfilename, "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                base_dir = {self.tempdir}
                step_length = 60
                albedo = a01.tif a02.tif a03.tif a04.tif a05.tif a06.tif
                         a07.tif a08.tif a09.tif a10.tif a11.tif a12.tif
                nighttime_solar_radiation_ratio = 0.8
                elevation = 8
                """
                ).format(self=self)
            )
        cli.App(self.configfilename).run()
        m.return_value.execute.assert_called_once_with()

    @patch("evaporation.cli.ProcessSpatial")
    def test_seasonal_albedo_configuration_as_mix_numbers_and_grids(self, m):
        with open(self.configfilename, "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                base_dir = {self.tempdir}
                step_length = 60
                albedo = 0.23 a02.tif a03.tif a04.tif a05.tif a06.tif
                         a07.tif a08.tif a09.tif a10.tif a11.tif a12.tif
                nighttime_solar_radiation_ratio = 0.8
                elevation = 8
                """
                ).format(self=self)
            )
        cli.App(self.configfilename).run()
        m.return_value.execute.assert_called_once_with()

    @patch("evaporation.cli.ProcessSpatial")
    def test_run_app_seasonal_albedo_with_float_sample_inputs(self, m):
        with open(self.configfilename, "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                base_dir = {self.tempdir}
                step_length = 60
                albedo = 0.10 0.23 0.34 0.24 0.45 0.46
                         0.34 0.12 0.14 0.78 0.78 0.12
                nighttime_solar_radiation_ratio = 0.8
                elevation = 8
                """
                ).format(self=self)
            )
        cli.App(self.configfilename).run()
        m.return_value.execute.assert_called_once_with()

    @patch("evaporation.cli.ProcessSpatial")
    def test_run_app_with_seasonal_albedo_with_grid_sample_inputs(self, m):
        with open(self.configfilename, "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                base_dir = {self.tempdir}
                step_length = 60
                albedo = a00.tif a00.tif a00.tif a00.tif a00.tif a00.tif
                         a00.tif a00.tif a00.tif a00.tif a00.tif a00.tif
                nighttime_solar_radiation_ratio = 0.8
                elevation = 8
                """
                ).format(self=self)
            )
        cli.App(self.configfilename).run()
        m.return_value.execute.assert_called_once_with()

    @patch("evaporation.cli.ProcessSpatial")
    def test_run_app_with_seasonal_albedo_with_mix_sample_inputs(self, m):
        with open(self.configfilename, "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                base_dir = {self.tempdir}
                step_length = 60
                albedo = a00.tif 0.23 a00.tif a00.tif a00.tif a00.tif
                         a00.tif a00.tif 0.23 a00.tif 0.23 a00.tif
                nighttime_solar_radiation_ratio = 0.8
                elevation = 8
                """
                ).format(self=self)
            )
        cli.App(self.configfilename).run()
        m.return_value.execute.assert_called_once_with()

    @patch("evaporation.cli.ProcessSpatial")
    def test_seasonal_albedo_configuration_as_12_numbers(self, m):
        with open(self.configfilename, "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                base_dir = {self.tempdir}
                step_length = 60
                albedo = 0.10 0.23 0.34 0.24 0.45 0.46
                         0.34 0.12 0.14 0.78 0.78 0.12
                nighttime_solar_radiation_ratio = 0.8
                elevation = 8
                """
                ).format(self=self)
            )
        cli.App(self.configfilename).run()
        m.return_value.execute.assert_called_once_with()


class CorrectConfigurationWithLogFileTestCase(TestCase):
    @patch("evaporation.cli.ProcessAtPoint")
    def test_creates_log_file(self, *args):
        with tempfile.TemporaryDirectory() as dirname:
            configfilename = os.path.join(dirname, "evaporation.conf")
            logfilename = os.path.join(dirname, "vaporize.log")
            with open(configfilename, "w") as f:
                f.write(
                    textwrap.dedent(
                        """\
                        base_dir = {}
                        albedo = 0.23
                        nighttime_solar_radiation_ratio = 0.8
                        step_length = 60
                        unit_converter_pressure = x / 10.0
                        unit_converter_solar_radiation = x * 3600 /1e6
                        logfile = {}
                        """
                    ).format(dirname, logfilename)
                )
            htsfilename = os.path.join(dirname, "somefile.hts")
            Path(htsfilename).touch()
            cli.App(configfilename).run()
            self.assertTrue(os.path.exists(logfilename))


class HtsTestCase(TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.config_file = os.path.join(self.tempdir, "vaporize.conf")
        self.savedcwd = os.getcwd()

    def tearDown(self):
        os.chdir(self.savedcwd)
        shutil.rmtree(self.tempdir)

    def setup_input_file(self, step, basename, value, missing=None):
        filename = os.path.join(self.tempdir, basename + ".hts")
        timestamp = step == "hourly" and "2014-10-01 15:00" or "2014-07-06"
        with open(filename, "w") as f:
            f.write("Title={}\n".format(basename.replace("_", " ").capitalize()))
            f.write("Timezone=CVT (UTC-0100)\n")
            if missing != "location":
                f.write("Location=-16.25 16.217 4326\n")
            if missing != "altitude":
                f.write("Altitude={}\n".format(step == "hourly" and "8" or "100"))
            f.write("\n{},{},\n".format(timestamp, value))

    def setup_hourly_input_files(self, missing=None):
        self.setup_input_file("hourly", "temperature", 38, missing=missing)
        self.setup_input_file("hourly", "humidity", 52, missing=missing)
        self.setup_input_file("hourly", "wind_speed", 3.3, missing=missing)
        self.setup_input_file("hourly", "pressure", 1013, missing=missing)
        self.setup_input_file("hourly", "solar_radiation", 681, missing=missing)

    def setup_daily_input_files(self):
        self.setup_input_file("daily", "temperature_max", 21.5)
        self.setup_input_file("daily", "temperature_min", 12.3)
        self.setup_input_file("daily", "humidity_max", 84)
        self.setup_input_file("daily", "humidity_min", 63)
        self.setup_input_file("daily", "wind_speed", 2.078)
        self.setup_input_file("daily", "sunshine_duration", 9.25)

    def setup_config_file(self, step_length):
        with open(self.config_file, "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                    base_dir = {self.tempdir}
                    albedo = 0.23
                    step_length = {step_length}
                    """
                ).format(self=self, step_length=step_length)
            )
            if step_length == 60:
                f.write(
                    textwrap.dedent(
                        """\
                        nighttime_solar_radiation_ratio = 0.8
                        unit_converter_pressure = x / 10.0
                        unit_converter_solar_radiation = x * 3600 / 1e6
                        """
                    )
                )

    def test_hourly(self):
        self.setup_hourly_input_files()
        self.setup_config_file(60)

        # Verify the output file doesn't exist yet
        result_filename = os.path.join(self.tempdir, "evaporation.hts")
        assert not os.path.exists(result_filename)

        # Execute
        cli.App(self.config_file).run()

        # Check that it has created a file and that the file is correct
        with open(result_filename) as f:
            t = HTimeseries(f)
        expected_result = pd.DataFrame(
            data={"value": [0.63], "flags": [""]},
            columns=("value", "flags"),
            index=[dt.datetime(2014, 10, 1, 15, 0)],
        )
        expected_result.index.name = "date"
        pd.testing.assert_frame_equal(t.data, expected_result, check_less_precise=2)

    def test_daily(self):
        self.setup_daily_input_files()
        self.setup_config_file(1440)

        # Verify the output file doesn't exist yet
        result_filename = os.path.join(self.tempdir, "evaporation.hts")
        assert not os.path.exists(result_filename)

        # Execute
        cli.App(self.config_file).run()

        # Check that it has created a file and that the file is correct
        with open(result_filename) as f:
            t = HTimeseries(f)
        expected_result = pd.DataFrame(
            data={"value": [3.9], "flags": [""]},
            columns=["value", "flags"],
            index=[dt.datetime(2014, 7, 6)],
        )
        expected_result.index.name = "date"
        pd.testing.assert_frame_equal(t.data, expected_result, check_less_precise=1)

    def test_missing_location(self):
        self.setup_hourly_input_files(missing="location")
        self.setup_config_file(60)
        msg = (
            "Incorrect or unspecified or inconsistent locations in the time series "
            "files."
        )
        with self.assertRaisesRegex(click.ClickException, msg):
            cli.App(self.config_file).run()

    def test_missing_altitude(self):
        self.setup_hourly_input_files(missing="altitude")
        self.setup_config_file(60)
        msg = (
            "Incorrect or unspecified or inconsistent altitudes in the time series "
            "files."
        )
        with self.assertRaisesRegex(click.ClickException, msg):
            cli.App(self.config_file).run()


class SpatialTestCase(TestCase):
    def setup_input_file(self, variable, value, with_date=True):
        """
        Saves value, which is an np array, to a GeoTIFF file whose name is
        based on variable.
        """
        if not with_date:
            filename = variable
        elif not isinstance(self.timestamp, dt.datetime):
            # Daily stuff
            filename = variable + "-" + self.timestamp.strftime("%Y-%m-%d")
        else:
            # Hourly stuff
            filename = variable + "-" + self.timestamp.strftime("%Y-%m-%d-%H-%M%z")

        filename = os.path.join(self.tempdir, filename + ".tif")
        f = gdal.GetDriverByName("GTiff").Create(filename, 2, 1, 1, gdal.GDT_Float32)

        nodata = -1.70141183e38

        value[np.isnan(value)] = nodata
        try:
            f.SetMetadataItem("TIMESTAMP", self.timestamp.isoformat())
            f.SetGeoTransform(self.geo_transform)
            f.SetProjection(self.wgs84.ExportToWkt())
            f.GetRasterBand(1).SetNoDataValue(nodata)
            f.GetRasterBand(1).WriteArray(value)
        finally:
            # Close the dataset
            f = None

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.config_file = os.path.join(self.tempdir, "vaporize.conf")
        self.savedcwd = os.getcwd()

        # Prepare data common to all input files
        self.geo_transform = (-16.25, 1.0, 0, 16.217, 0, 1.0)
        self.wgs84 = osr.SpatialReference()
        self.wgs84.ImportFromEPSG(4326)

        # Prepare sample albedo grid
        filename = os.path.join(self.tempdir, "a00.tif")
        create_geotiff_file(filename, np.array([[0.23, 0.44]]))

        # Save standard error (some tests change it)
        self.saved_stderr = sys.stderr

    def tearDown(self):
        os.chdir(self.savedcwd)
        shutil.rmtree(self.tempdir)

    def test_execute_notz(self):
        # Prepare input files without time zone
        self.timestamp = dt.datetime(2014, 10, 1, 15, 0)
        self.setup_input_file("temperature-notz", np.array([[38.0, 28.0]]))
        self.setup_input_file("humidity-notz", np.array([[52.0, 42.0]]))
        self.setup_input_file("wind_speed-notz", np.array([[3.3, 2.3]]))
        self.setup_input_file("pressure-notz", np.array([[1013.0, 1013.0]]))
        self.setup_input_file("solar_radiation-notz", np.array([[681.0, 403.0]]))

        with open(self.config_file, "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                base_dir = {self.tempdir}
                albedo = 0.23
                nighttime_solar_radiation_ratio = 0.8
                elevation = 8
                step_length = 60
                unit_converter_pressure = x / 10.0
                unit_converter_solar_radiation = x * 3600 / 1e6
                temperature_prefix = temperature-notz
                humidity_prefix = humidity-notz
                wind_speed_prefix = wind_speed-notz
                solar_radiation_prefix = solar_radiation-notz
                pressure_prefix = pressure-notz
                """
                ).format(self=self)
            )

        # Verify the output file doesn't exist yet
        result_filename = os.path.join(
            self.tempdir,
            "evaporation-{}.tif".format(self.timestamp.strftime("%Y-%m-%d-%H-%M%z")),
        )
        self.assertFalse(os.path.exists(result_filename))

        # Execute
        with self.assertRaisesRegex(Exception, "time zone"):
            cli.App(self.config_file).run()

        # Verify the output file still doesn't exist
        self.assertFalse(os.path.exists(result_filename))

    def test_execute_daily(self):
        # Prepare input files
        self.timestamp = dt.date(2014, 7, 6)
        self.setup_input_file("temperature_max", np.array([[21.5, 28]]))
        self.setup_input_file("temperature_min", np.array([[12.3, 15]]))
        self.setup_input_file("humidity_max", np.array([[84.0, 70.0]]))
        self.setup_input_file("humidity_min", np.array([[63.0, 60.0]]))
        self.setup_input_file("wind_speed", np.array([[2.078, 2.244]]))
        self.setup_input_file("sunshine_duration", np.array([[9.25, 9.0]]))

        # Also setup an output file that has no corresponding input files
        rogue_output_file = os.path.join(self.tempdir, "evaporation-2013-01-01.tif")
        with open(rogue_output_file, "w") as f:
            f.write("irrelevant contents")

        with open(self.config_file, "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                base_dir = {self.tempdir}
                albedo = 0.23
                elevation = 100
                step_length = 1440
                """
                ).format(self=self)
            )

        # Verify the output file doesn't exist yet
        result_filename = os.path.join(
            self.tempdir,
            "evaporation-{}.tif".format(self.timestamp.strftime("%Y-%m-%d")),
        )
        self.assertFalse(os.path.exists(result_filename))

        # Verify the rogue output file is still here
        self.assertTrue(os.path.exists(rogue_output_file))

        # Execute
        cli.App(self.config_file).run()

        # Check that it has created a file
        self.assertTrue(os.path.exists(result_filename))

        # Check that the rogue output file is gone
        self.assertFalse(os.path.exists(rogue_output_file))

        # Check that the created file is correct
        fp = gdal.Open(result_filename)
        timestamp = fp.GetMetadata()["TIMESTAMP"]
        self.assertEqual(timestamp, "2014-07-06")
        self.assertEqual(fp.RasterXSize, 2)
        self.assertEqual(fp.RasterYSize, 1)
        self.assertEqual(fp.GetGeoTransform(), self.geo_transform)
        # We can't just compare fp.GetProjection() to self.wgs84.ExportToWkt(),
        # because sometimes there are minor differences in the formatting or in
        # the information contained in the WKT.
        self.assertTrue(fp.GetProjection().startswith('GEOGCS["WGS 84",'))
        self.assertTrue(fp.GetProjection().endswith('AUTHORITY["EPSG","4326"]]'))
        np.testing.assert_almost_equal(
            fp.GetRasterBand(1).ReadAsArray(), np.array([[3.9, 4.8]]), decimal=1
        )
        fp = None

    def test_execute_daily_with_radiation(self):
        """Same as test_execute_daily, except that we use solar radiation
           instead of sunshine duration."""
        # Prepare input files
        self.timestamp = dt.date(2014, 7, 6)
        self.setup_input_file("temperature_max", np.array([[21.5, 28]]))
        self.setup_input_file("temperature_min", np.array([[12.3, 15]]))
        self.setup_input_file("humidity_max", np.array([[84.0, 70.0]]))
        self.setup_input_file("humidity_min", np.array([[63.0, 60.0]]))
        self.setup_input_file("wind_speed", np.array([[2.078, 2.244]]))
        self.setup_input_file("solar_radiation", np.array([[22.07, 21.62]]))

        # Also setup an output file that has no corresponding input files
        rogue_output_file = os.path.join(self.tempdir, "evaporation-2013-01-01.tif")
        with open(rogue_output_file, "w") as f:
            f.write("irrelevant contents")

        with open(self.config_file, "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                base_dir = {self.tempdir}
                albedo = 0.23
                elevation = 100
                step_length = 1440
                """
                ).format(self=self)
            )

        # Verify the output file doesn't exist yet
        result_filename = os.path.join(
            self.tempdir,
            "evaporation-{}.tif".format(self.timestamp.strftime("%Y-%m-%d")),
        )
        self.assertFalse(os.path.exists(result_filename))

        # Verify the rogue output file is still here
        self.assertTrue(os.path.exists(rogue_output_file))

        # Execute
        cli.App(self.config_file).run()

        # Check that it has created a file
        self.assertTrue(os.path.exists(result_filename))

        # Check that the rogue output file is gone
        self.assertFalse(os.path.exists(rogue_output_file))

        # Check that the created file is correct
        fp = gdal.Open(result_filename)
        timestamp = fp.GetMetadata()["TIMESTAMP"]
        self.assertEqual(timestamp, "2014-07-06")
        self.assertEqual(fp.RasterXSize, 2)
        self.assertEqual(fp.RasterYSize, 1)
        self.assertEqual(fp.GetGeoTransform(), self.geo_transform)
        # We can't just compare fp.GetProjection() to self.wgs84.ExportToWkt(),
        # because sometimes there are minor differences in the formatting or in
        # the information contained in the WKT.
        self.assertTrue(fp.GetProjection().startswith('GEOGCS["WGS 84",'))
        self.assertTrue(fp.GetProjection().endswith('AUTHORITY["EPSG","4326"]]'))
        np.testing.assert_almost_equal(
            fp.GetRasterBand(1).ReadAsArray(), np.array([[3.9, 4.8]]), decimal=1
        )
        fp = None

    def test_execute_hourly(self):
        # Prepare input files
        self.timestamp = dt.datetime(2014, 10, 1, 15, 0, tzinfo=senegal_tzinfo)
        self.setup_input_file("temperature", np.array([[38.0, 28.0]]))
        self.setup_input_file("humidity", np.array([[52.0, 42.0]]))
        self.setup_input_file("wind_speed", np.array([[3.3, 2.3]]))
        self.setup_input_file("pressure", np.array([[1013.0, 1013.0]]))
        self.setup_input_file("solar_radiation", np.array([[681.0, 403.0]]))

        # Also setup an output file that has no corresponding input files
        rogue_output_file = os.path.join(
            self.tempdir, "evaporation-2013-01-01-15-00-0100.tif"
        )
        with open(rogue_output_file, "w") as f:
            f.write("irrelevant contents")

        with open(self.config_file, "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                base_dir = {self.tempdir}
                albedo = 0.23
                nighttime_solar_radiation_ratio = 0.8
                elevation = 8
                step_length = 60
                unit_converter_pressure = x / 10.0
                unit_converter_solar_radiation = x * 3600 / 1e6
                """
                ).format(self=self)
            )

        # Verify the output file doesn't exist yet
        result_filename = os.path.join(
            self.tempdir,
            "evaporation-{}.tif".format(self.timestamp.strftime("%Y-%m-%d-%H-%M%z")),
        )
        self.assertFalse(os.path.exists(result_filename))

        # Verify the rogue output file is still here
        self.assertTrue(os.path.exists(rogue_output_file))

        # Execute
        cli.App(self.config_file).run()

        # Check that it has created a file
        self.assertTrue(os.path.exists(result_filename))

        # Check that the rogue output file is gone
        self.assertFalse(os.path.exists(rogue_output_file))

        # Check that the created file is correct
        fp = gdal.Open(result_filename)
        timestamp = fp.GetMetadata()["TIMESTAMP"]
        self.assertEqual(timestamp, "2014-10-01T15:00:00-01:00")
        self.assertEqual(fp.RasterXSize, 2)
        self.assertEqual(fp.RasterYSize, 1)
        self.assertEqual(fp.GetGeoTransform(), self.geo_transform)
        # We can't just compare fp.GetProjection() to self.wgs84.ExportToWkt(),
        # because sometimes there are minor differences in the formatting or in
        # the information contained in the WKT.
        self.assertTrue(fp.GetProjection().startswith('GEOGCS["WGS 84",'))
        self.assertTrue(fp.GetProjection().endswith('AUTHORITY["EPSG","4326"]]'))
        np.testing.assert_almost_equal(
            fp.GetRasterBand(1).ReadAsArray(), np.array([[0.63, 0.36]]), decimal=2
        )
        fp = None

    def test_execute_hourly_no_pressure(self):
        """Same as test_execute_hourly, but does not have pressure an input;
           therefore, it will calculate pressure itself."""
        # Prepare input files
        self.timestamp = dt.datetime(2014, 10, 1, 15, 0, tzinfo=senegal_tzinfo)
        self.setup_input_file("temperature", np.array([[38.0, 28.0]]))
        self.setup_input_file("humidity", np.array([[52.0, 42.0]]))
        self.setup_input_file("wind_speed", np.array([[3.3, 2.3]]))
        self.setup_input_file("solar_radiation", np.array([[681.0, 403.0]]))

        # Also setup an output file that has no corresponding input files
        rogue_output_file = os.path.join(
            self.tempdir, "evaporation-2013-01-01-15-00-0100.tif"
        )
        with open(rogue_output_file, "w") as f:
            f.write("irrelevant contents")

        with open(self.config_file, "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                base_dir = {self.tempdir}
                albedo = 0.23
                nighttime_solar_radiation_ratio = 0.8
                elevation = 8
                step_length = 60
                unit_converter_solar_radiation = x * 3600 / 1e6
                """
                ).format(self=self)
            )

        # Verify the output file doesn't exist yet
        result_filename = os.path.join(
            self.tempdir,
            "evaporation-{}.tif".format(self.timestamp.strftime("%Y-%m-%d-%H-%M%z")),
        )
        self.assertFalse(os.path.exists(result_filename))

        # Verify the rogue output file is still here
        self.assertTrue(os.path.exists(rogue_output_file))

        # Execute
        cli.App(self.config_file).run()

        # Check that it has created a file
        self.assertTrue(os.path.exists(result_filename))

        # Check that the rogue output file is gone
        self.assertFalse(os.path.exists(rogue_output_file))

        # Check that the created file is correct
        fp = gdal.Open(result_filename)
        timestamp = fp.GetMetadata()["TIMESTAMP"]
        self.assertEqual(timestamp, "2014-10-01T15:00:00-01:00")
        self.assertEqual(fp.RasterXSize, 2)
        self.assertEqual(fp.RasterYSize, 1)
        self.assertEqual(fp.GetGeoTransform(), self.geo_transform)
        # We can't just compare fp.GetProjection() to self.wgs84.ExportToWkt(),
        # because sometimes there are minor differences in the formatting or in
        # the information contained in the WKT.
        self.assertTrue(fp.GetProjection().startswith('GEOGCS["WGS 84",'))
        self.assertTrue(fp.GetProjection().endswith('AUTHORITY["EPSG","4326"]]'))
        np.testing.assert_almost_equal(
            fp.GetRasterBand(1).ReadAsArray(), np.array([[0.63, 0.36]]), decimal=2
        )
        fp = None

    def test_execute_hourly_without_sun(self):
        # Prepare input files, without solar radiation
        self.timestamp = dt.datetime(2014, 10, 1, 15, 0, tzinfo=senegal_tzinfo)
        self.setup_input_file("temperature", np.array([[38.0, 28.0]]))
        self.setup_input_file("humidity", np.array([[52.0, 42.0]]))
        self.setup_input_file("wind_speed", np.array([[3.3, 2.3]]))
        self.setup_input_file("pressure", np.array([[1013.0, 1013.0]]))

        # Configuration
        with open(self.config_file, "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                base_dir = {self.tempdir}
                albedo = 0.23
                nighttime_solar_radiation_ratio = 0.8
                elevation = 8
                step_length = 60
                unit_converter_pressure = x / 10.0
                unit_converter_solar_radiation = x * 3600 / 1e6
                """
                ).format(self=self)
            )

        # Execute and check exception
        msg = "Neither sunshine_duration nor solar_radiation are available"
        with self.assertRaisesRegex(click.ClickException, msg):
            cli.App(self.config_file).run()

    def test_execute_with_dem(self):
        """This is essentially the same as test_execute, but uses a GeoTIFF
        with a DEM instead of a constant elevation. The numbers are the same,
        however (all DEM gridpoints have the same value)."""

        # Prepare input files
        self.timestamp = dt.datetime(2014, 10, 1, 15, 0, tzinfo=senegal_tzinfo)
        self.setup_input_file("temperature", np.array([[38.0, 28.0]]))
        self.setup_input_file("humidity", np.array([[52.0, 42.0]]))
        self.setup_input_file("wind_speed", np.array([[3.3, 2.3]]))
        self.setup_input_file("pressure", np.array([[1013.0, 1013.0]]))
        self.setup_input_file("solar_radiation", np.array([[681.0, 403.0]]))
        self.setup_input_file("dem", np.array([[8.0, 8.0]]), with_date=False)

        with open(self.config_file, "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                base_dir = {self.tempdir}
                albedo = 0.23
                nighttime_solar_radiation_ratio = 0.8
                elevation = dem.tif
                step_length = 60
                unit_converter_pressure = x / 10.0
                unit_converter_solar_radiation = x * 3600 / 1e6
                """
                ).format(self=self)
            )

        # Verify the output file doesn't exist yet
        result_filename = os.path.join(
            self.tempdir,
            "evaporation-{}.tif".format(self.timestamp.strftime("%Y-%m-%d-%H-%M%z")),
        )
        self.assertFalse(os.path.exists(result_filename))

        # Execute
        cli.App(self.config_file).run()

        # Check that it has created a file
        self.assertTrue(os.path.exists(result_filename))

        # Check that the created file is correct
        fp = gdal.Open(result_filename)
        timestamp = fp.GetMetadata()["TIMESTAMP"]
        self.assertEqual(timestamp, "2014-10-01T15:00:00-01:00")
        self.assertEqual(fp.RasterXSize, 2)
        self.assertEqual(fp.RasterYSize, 1)
        self.assertEqual(fp.GetGeoTransform(), self.geo_transform)
        # We can't just compare fp.GetProjection() to self.wgs84.ExportToWkt(),
        # because sometimes there are minor differences in the formatting or in
        # the information contained in the WKT.
        self.assertTrue(fp.GetProjection().startswith('GEOGCS["WGS 84",'))
        self.assertTrue(fp.GetProjection().endswith('AUTHORITY["EPSG","4326"]]'))
        np.testing.assert_almost_equal(
            fp.GetRasterBand(1).ReadAsArray(), np.array([[0.63, 0.36]]), decimal=2
        )
        fp = None

    def test_execute_with_nodata(self):
        """This is essentially the same as test_execute, but the gdal rasters
        contain cells with nodata."""

        sys.stderr = StringIO()

        # Prepare input files
        self.timestamp = dt.datetime(2014, 10, 1, 15, 0, tzinfo=senegal_tzinfo)
        nan = float("nan")
        self.setup_input_file("temperature", np.array([[38.0, nan]]))
        self.setup_input_file("humidity", np.array([[52.0, nan]]))
        self.setup_input_file("wind_speed", np.array([[3.3, 2.3]]))
        self.setup_input_file("pressure", np.array([[1013.0, nan]]))
        self.setup_input_file("solar_radiation", np.ma.array([[681.0, nan]]))
        self.setup_input_file("dem", np.array([[8.0, nan]]), with_date=False)

        with open(self.config_file, "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                base_dir = {self.tempdir}
                albedo = 0.23
                nighttime_solar_radiation_ratio = 0.8
                elevation = dem.tif
                step_length = 60
                unit_converter_pressure = x / 10.0
                unit_converter_solar_radiation = x * 3600 / 1e6
                """
                ).format(self=self)
            )

        # Verify the output file doesn't exist yet
        result_filename = os.path.join(
            self.tempdir,
            "evaporation-{}.tif".format(self.timestamp.strftime("%Y-%m-%d-%H-%M%z")),
        )
        self.assertFalse(os.path.exists(result_filename))

        # Execute
        cli.App(self.config_file).run()

        # Check that it has created a file
        self.assertTrue(os.path.exists(result_filename))

        # Check that the created file is correct
        fp = gdal.Open(result_filename)
        timestamp = fp.GetMetadata()["TIMESTAMP"]
        self.assertEqual(timestamp, "2014-10-01T15:00:00-01:00")
        self.assertEqual(fp.RasterXSize, 2)
        self.assertEqual(fp.RasterYSize, 1)
        self.assertEqual(fp.GetGeoTransform(), self.geo_transform)
        # We can't just compare fp.GetProjection() to self.wgs84.ExportToWkt(),
        # because sometimes there are minor differences in the formatting or in
        # the information contained in the WKT.
        self.assertTrue(fp.GetProjection().startswith('GEOGCS["WGS 84",'))
        self.assertTrue(fp.GetProjection().endswith('AUTHORITY["EPSG","4326"]]'))
        nodatavalue = fp.GetRasterBand(1).GetNoDataValue()
        np.testing.assert_almost_equal(
            fp.GetRasterBand(1).ReadAsArray(),
            np.array([[0.63, nodatavalue]]),
            decimal=2,
        )

        # sys.stderr should be empty (no RuntimeWarning should have been
        # written there)
        self.assertEqual("", sys.stderr.getvalue())

        fp = None
